-- This file is automatically generated using maintenance/generateSchemaChangeSql.php.
-- Source: schema/abstractSchemaChanges/patch-cu_changes-drop-cuc_user.json
-- Do not modify this file directly.
-- See https://www.mediawiki.org/wiki/Manual:Schema_changes
DROP INDEX cuc_user_ip_time ON /*_*/cu_changes;

ALTER TABLE /*_*/cu_changes
  DROP cuc_user,
  DROP cuc_user_text;
