<?php
/**
 * @license GPL-2.0-or-later
 * @file
 */

namespace MediaWiki\Storage;

use RuntimeException;

/**
 * Exception representing a failure to update a page entry.
 *
 * @newable
 * @since 1.32
 */
class PageUpdateException extends RuntimeException {

}
