<?php
/**
 * Database row sorting.
 *
 * @license GPL-2.0-or-later
 * @file
 */

/**
 * @since 1.16.3
 * @author Tim Starling
 * @stable to extend
 */
abstract class Collation {

	/**
	 * Given a string, convert it to a (hopefully short) key that can be used
	 * for efficient sorting.  A binary sort according to the sortkeys
	 * corresponds to a logical sort of the corresponding strings.  Current
	 * code expects that a line feed character should sort before all others, but
	 * has no other particular expectations (and that one can be changed if
	 * necessary).
	 *
	 * @since 1.16.3
	 *
	 * @param string $string UTF-8 string
	 * @return string Binary sortkey
	 */
	abstract public function getSortKey( $string );

	/**
	 * Get multiple sort keys
	 *
	 * @param string[] $strings
	 * @return string[]
	 */
	public function getSortKeys( $strings ) {
		$ret = [];
		foreach ( $strings as $key => $s ) {
			$ret[$key] = $this->getSortKey( $s );
		}
		return $ret;
	}

	/**
	 * Given a string, return the logical "first letter" to be used for
	 * grouping on category pages and so on.  This has to be coordinated
	 * carefully with convertToSortkey(), or else the sorted list might jump
	 * back and forth between the same "initial letters" or other pathological
	 * behavior.  For instance, if you just return the first character, but "a"
	 * sorts the same as "A" based on getSortKey(), then you might get a
	 * list like
	 *
	 * == A ==
	 * * [[Aardvark]]
	 *
	 * == a ==
	 * * [[antelope]]
	 *
	 * == A ==
	 * * [[Ape]]
	 *
	 * etc., assuming for the sake of argument that $wgCapitalLinks is false.
	 *
	 * @since 1.16.3
	 *
	 * @param string $string UTF-8 string
	 * @return string UTF-8 string corresponding to the first letter of input
	 */
	abstract public function getFirstLetter( $string );

}
