/*!
 * VisualEditor ContentEditable MediaWiki MWTransclusionNode styles.
 *
 * @copyright See AUTHORS.txt
 * @license The MIT License (MIT); see LICENSE.txt
 */

.ve-ce-mwTransclusionNode {
	> .ve-ce-focusableNode-invisibleIcon.oo-ui-iconElement {
		font-size: 1em;
		width: auto;
	}

	&.ve-ce-focusableNode-invisible {
		width: auto;
	}
}
