/*!
 * OOUI v0.53.0
 * https://www.mediawiki.org/wiki/OOUI
 *
 * Copyright 2011–2025 OOUI Team and other contributors.
 * Released under the MIT license
 * http://oojs.mit-license.org
 *
 * Date: 2025-09-03T22:03:24Z
 */
.oo-ui-icon-alert {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/alert.svg);
}
.oo-ui-icon-bell {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/bell.svg);
}
.oo-ui-icon-bellOutline {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/bellOutline.svg);
}
.oo-ui-icon-error {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/error.svg);
}
.oo-ui-icon-message {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/message.svg);
}
.oo-ui-icon-notice {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/notice.svg);
}
.oo-ui-icon-speechBubble {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/speechBubble-rtl.svg);
}
.oo-ui-icon-speechBubbleAdd {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/speechBubbleAdd-rtl.svg);
}
.oo-ui-icon-speechBubbles {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/speechBubbles-rtl.svg);
}
.oo-ui-icon-success {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/success.svg);
}
.oo-ui-icon-tray {
  background-image: /* @embed */ url(themes/wikimediaui/images/icons/tray.svg);
}
