@import 'mediawiki.skin.variables.less';

.ext-confirmEdit-hCaptchaLoadingIndicator {
	display: flex;
	justify-content: center;
	margin: 0 auto;
	padding-bottom: @spacing-100;
}

.ve-ui-mwSaveDialog-savePanel {
	.hcaptcha-widget-container-hide-icon {
		.oo-ui-iconElement-icon {
			display: none;
		}

		.oo-ui-labelElement-label {
			margin-left: 0;
		}
	}
}
