<div id="wikieditor-toolbar-replace-message">
	<div id="wikieditor-toolbar-replace-nomatch" rel="wikieditor-toolbar-tool-replace-nomatch"></div>
	<div id="wikieditor-toolbar-replace-success"></div>
	<div id="wikieditor-toolbar-replace-emptysearch" rel="wikieditor-toolbar-tool-replace-emptysearch"></div>
	<div id="wikieditor-toolbar-replace-invalidregex"></div>
</div>
<fieldset>
	<div class="wikieditor-toolbar-field-wrapper">
		<label for="wikieditor-toolbar-replace-search" rel="wikieditor-toolbar-tool-replace-search"></label>
		<input type="text" id="wikieditor-toolbar-replace-search"/>
	</div>
	<div class="wikieditor-toolbar-field-wrapper">
		<label for="wikieditor-toolbar-replace-replace" rel="wikieditor-toolbar-tool-replace-replace"></label>
		<input type="text" id="wikieditor-toolbar-replace-replace"/>
	</div>
	<div class="wikieditor-toolbar-field-wrapper">
		<input type="checkbox" id="wikieditor-toolbar-replace-case"/>
		<label for="wikieditor-toolbar-replace-case" rel="wikieditor-toolbar-tool-replace-case"></label>
	</div>
	<div class="wikieditor-toolbar-field-wrapper">
		<input type="checkbox" id="wikieditor-toolbar-replace-word"/>
		<label for="wikieditor-toolbar-replace-word" rel="wikieditor-toolbar-tool-replace-word"></label>
	</div>
	<div class="wikieditor-toolbar-field-wrapper">
		<input type="checkbox" id="wikieditor-toolbar-replace-regex"/>
		<label for="wikieditor-toolbar-replace-regex" rel="wikieditor-toolbar-tool-replace-regex"></label>
	</div>
</fieldset>
