<?php

namespace MediaWiki\ResourceLoader;

use RuntimeException;

/**
 * Exception thrown when a network error occurs while fetching a foreign resource.
 */
class ForeignResourceNetworkException extends RuntimeException {
}
